// BlinkenSisters - Hunt for the Lost Pixels
//     Bringing back the fun of the 80s
//
// (C) 2005-07 Rene Schickbauer, Wolfgang Dautermann
//
// See License.txt for licensing information
//


#include "globals.h"
#include "bl_lua.h"
#include "errorhandler.h"
#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"
#include "ls_luasocket.h"
#include "bl_lua_bindings.h"
#include "bl_lua_objbindings.h"
#include "ls_luasocket.h"
#include "bl_lua_obj_fgobjects.h"
#include "bl_lua_obj_timer.h"
#include "bl_lua_obj_gfx.h"
#include "bl_lua_obj_anim.h"

#include <stdarg.h>
#include <string.h>

lua_State* scriptIsRunning;

void blLuaInit(char *fname, bool OOBindings)
{
	lua_State* blLUA;
	char fullfname[MAX_FNAME_LENGTH];
	sprintf(fullfname, "%s", configGetPath(fname));

	// Initialize LUA
	blLUA = lua_open();
	if(!blLUA) {
		DIE(ERROR_LUAINIT, "");
	}

	// Load libraries
	luaopen_base(blLUA);
	luaopen_io(blLUA);
	luaopen_string(blLUA);
	luaopen_math(blLUA);
	luaopen_socket_core(blLUA);
	if(!OOBindings) {
		blRegisterLuaBindings(blLUA);
	} else {
		blRegisterLuaObjBindings(blLUA);
	}

	scriptIsRunning = blLUA;
	// Load script file
	if(luaL_loadfile(blLUA, fullfname)) {
		DIE(ERROR_LUALOAD, (char *)lua_tostring(blLUA, -1));
	}
	
	// Parse script file (for globals and such)
	if(lua_pcall(blLUA,0,0,0)) {
		DIE(ERROR_LUAPARSE, (char *)lua_tostring(blLUA, -1));
	}
	
	if(!OOBindings) {
		lhandle.blLuaState = blLUA;
	} else {
		lhandle.blOOLuaState = blLUA;
	}
	
	scriptIsRunning = 0;
}

void blLuaDeInit()
{
	if(lhandle.blLuaState) {
		lua_close(lhandle.blLuaState);
		lhandle.blLuaState = NULL;
	}
	if(lhandle.blOOLuaState) {
		lua_close(lhandle.blOOLuaState);
		lhandle.blOOLuaState = NULL;
	}
}

void blLuaCall(lua_State* blState, const char *func, const char *sig, ...)
{
	if(!blState) {
		DIE(ERROR_LUANOINIT, "");
	}
	
	va_list vl;
	int narg, nres, npopres;
	
	va_start(vl, sig);
	lua_getglobal(blState, func);
	
	// push arguments
	narg = 0;
	bool deli = false;
	while(*sig && !deli) {
		switch(*sig) {
			case 'd': // Double
				lua_pushnumber(blState, va_arg(vl, double));
				narg++;
				break;
			case 'i': // Int
				lua_pushnumber(blState, va_arg(vl, int));
				narg++;
				break;
			case 's': // String
				lua_pushstring(blState, va_arg(vl, char *));
				narg++;
				break;
			case 'F': // OOLUA FG Object id (push the whole object
				luaObjPushFGObject(blState, (Uint32)va_arg(vl, Uint32));
				narg++;
				break;
			case 'C': // OOLUA Colission / fgOverlap table (must be pre-filled from last col. check)
				luaObjPushFGOverlapObject(blState);
				narg++;
				break;
			case 'T': // OOLUA TIMER Object id (push the whole object
				luaObjPushTimer(blState, (Uint32)va_arg(vl, Uint32));
				narg++;
				break;
			case 'G': // OOLUA GFX Object id (push the whole object
				luaObjPushGraphic(blState, (Uint32)va_arg(vl, Uint32));
				narg++;
				break;				
			case 'A': // OOLUA GFX Object id (push the whole object
				luaObjPushAnim(blState, (Uint32)va_arg(vl, Uint32));
				narg++;
				break;
			case '*': // Pre-pushed object for OO Lua-Bindings, must at the start of the string!
				narg++;
				break;
			case '>': // Delimeter for results
				deli = true;
				break;
			default:
				char tmp[20];
				sprintf(tmp, "FORMAT: (%c)", *sig);
				DIE(ERROR_LUAPARAMFORMAT, tmp);
		}
		luaL_checkstack(blState, 1, "too many arguments");
		sig++;
	}
	
	// call the function
	scriptIsRunning = blState;
	npopres = nres = strlen(sig); // number of expected results
	if(lua_pcall(blState, narg, nres, 0)) {
		char tmp[200];
		sprintf(tmp, "Function: '%s'\n%s", func, lua_tostring(blState, -1));
		DIE(ERROR_LUACALL, tmp);
	}
	scriptIsRunning = 0;
	
	// Get results
	nres = -nres; // Stack index of first result
	while(*sig) {
		switch(*sig) {
			case 'd': // double
				if(!lua_isnumber(blState, nres)) {
					DIE(ERROR_LUARESULTTYPE, func);
				}
				*va_arg(vl, double *) = lua_tonumber(blState, nres);
				break;
			case 'i': // int
				if(!lua_isnumber(blState, nres)) {
					DIE(ERROR_LUARESULTTYPE, func);
				}
				*va_arg(vl, int *) = (int)lua_tonumber(blState, nres);
				break;
			case 's': // string
				if(!lua_isstring(blState, nres)) {
					DIE(ERROR_LUARESULTTYPE, func);
				}
				sprintf(*va_arg(vl, char **), "%s", lua_tostring(blState, nres));
				break;
			default:
				char tmp[20];
				sprintf(tmp, "FORMAT: (%c)", *sig);
				DIE(ERROR_LUAPARAMFORMAT, tmp);
		}
		sig++;
	}
	
	va_end(vl);

	while(npopres) {
		lua_pop(blState, 1);
		npopres--;
	}
	
}

void blLuaDieWithError(Uint32 errorcode, const char* extrainfo, Uint32 linenum, const char* filename) {
	//	fprintf(stderr, "Error %d: %s\nError-Info: %s\nError at %s:%d\n",
	//errorcode, getErrorText(errorcode), extrainfo, filename, linenum);
	
	char errorforlua[2000];
	sprintf(errorforlua, "\nCall-Error %d: %s\nCall-Error additional Info: %s\n   at %s:%d",
			errorcode, getErrorText(errorcode), extrainfo, filename, linenum);
	luaL_error(scriptIsRunning, errorforlua);
	
}
